import re
from struct import pack
from typing import Generator


class TensorData:
    """
    Match tensor data structure in Edge Impulse model.
    Patch format is (for each tensor):
        index: 1 byte
        bits: 1 byte (either 8 or 32)
        size: 2 bytes
        data: size * (bits // 8) bytes
    """
    def __init__(self, contents: str):
        """

        :param contents:
        """
        def eval_data(data: str) -> list:
            return [int(x) for x in eval("[" + re.sub(r'/\*.+?\*/', "", data) + "]")]

        tensors = re.findall(r'int(32|8)_t tensor_data(\d+)\[([0-9*]+)] = \{([\s\S]+?)};', contents)
        assert len(tensors) > 0, "Can't find tensor data"
        self.tensors = [(int(index), int(bits), eval(size), eval_data(data)) for bits, index, size, data in tensors]

    @property
    def byte_size(self) -> int:
        return len(self.bytes)

    @property
    def bytes(self) -> bytes:
        """
        Convert data to binary format
        :return:
        """
        packs = []

        for index, bits, size, data in self.tensors:
            assert bits in [8, 32], f'unknown bit depth for tensor {index}: {bits}'
            dtype = 'i' if bits == 32 else 'b'
            packs.append(pack(f">BBH{dtype * len(data)}", index, bits, size, *data))

        return b"".join(packs)

    @property
    def patches(self) -> Generator:
        """
        Generate code to patch tensor data
        :return:
        """
        for index, bits, size, data in self.tensors:
            yield f"""if (!ei_patch_tensor_data<int{bits}_t>(stream, {index}, {bits}, {size})) return false;"""



import re
import requests
from io import BytesIO
from typing import Union
from zipfile import ZipFile, ZIP_DEFLATED


class InputLibrary:
    """
    Read an Edge Impulse library from many sources:
     - raw zip bytes
     - local file
     - API
    """
    def __init__(self, zip: str | bytes = None, api_key: str = None, project_id: str = None, model_type: str = "int8", engine: str = "tflite-eon"):
        """
        Constructor
        :param zip: if str, it is interpreted as a path to a local file. If bytes, it is interpreted as the zip contents
        :param api_key: Edge Impulse API key
        :param project_id: Edge Impulse project ID
        :param model_type: int8 | float32
        :param engine: tflite | tflite-eon | tflite-eon-ram-optimized
        """
        self.bytes = None

        if isinstance(zip, str):
            self.bytes = self.from_local(zip)
        elif isinstance(zip, bytes):
            self.bytes = zip
        else:
            assert len(api_key) >= 50 and len(project_id) >= 6, "Missing either zip source or api_key/project_id"
            assert model_type in ["int8", "float32"], f"Unknown model type {model_type}"
            assert engine in ["tflite", "tflite-eon", "tflite-eon-ram-optimized"], f"Unknown engine {engine}"
            self.bytes = self.fetch(api_key, project_id, engine=engine, model_type=model_type)

    def from_local(self, path: str) -> bytes:
        """
        Read local zip file
        :param path:
        :return:
        """
        with open(path, "rb") as f:
            return f.read()

    def fetch(self, api_key: str, project_id: str, model_type: str, engine: str):
        """
        Fetch library from API
        :param api_key:
        :param project_id:
        :param model_type: #todo not used for now
        :param engine:
        :return:
        """
        content = b""
        url = f"https://studio.edgeimpulse.com/v1/api/{project_id}/deployment/download?type=arduino&engine={engine}"
        res = requests.get(url, headers={"X-Api-Key": api_key}, stream=True)
        assert res.status_code == 200, f"Bad response: {res.text}"

        # assert model is not larger than 10 MB
        for chunk in res.iter_content(chunk_size=1024 * 1024):
            content += chunk
            assert len(content) < 10 * 1024 * 1024, f"Model is too large: >10 MB"

        return content

    def replace(self, replacements: dict[str, Union[str, callable]]):
        """
        Replace contents of files
        :param replacements: {file pattern: replacement function}
        :return:
        """
        dest_io = BytesIO()

        def matches(filename: str, pattern: str) -> bool:
            """
            Check if filename matches glob-like pattern
            :param filename:
            :param pattern:
            :return:
            """
            return re.search(pattern.replace("*", ".*"), filename) is not None

        with ZipFile(BytesIO(self.bytes), "r") as zip:
            with ZipFile(dest_io, "w", ZIP_DEFLATED) as dest:
                for info in zip.infolist():
                    filename = info.filename
                    contents = zip.read(filename)

                    for pattern, replacement_function in replacements.items():
                        if matches(filename, pattern):
                            contents = replacement_function(contents.decode())
                            break

                    dest.writestr(info, contents)

        self.bytes = dest_io.getvalue()



import struct
import zipfile
from io import BytesIO

from edgeimpulse_ota.InputLibrary import InputLibrary
from edgeimpulse_ota.TensorData import TensorData
from edgeimpulse_ota.Quantization import Quantization


def serve(zip: str | bytes = None, api_key: str = None, project_id: str = None, **kwargs):
    """
    Convert model's weights to OTA patch
    :param zip: if str, it is interpreted as a path to a local file. If bytes, it is interpreted as the zip contents
    :param api_key: Edge Impulse API key
    :param project_id: Edge Impulse project ID
    :return:
    """
    library = InputLibrary(zip=zip, api_key=api_key, project_id=project_id, **kwargs)

    with zipfile.ZipFile(BytesIO(library.bytes), "r") as zip:
        for info in zip.infolist():
            filename = info.filename

            if "/tflite-model/tflite_learn_" in filename and filename.endswith(".cpp"):
                with zip.open(filename) as file:
                    contents = file.read().decode("utf-8")
                    tensor_data = TensorData(contents)
                    quantization = Quantization(contents)
                    patch_size = tensor_data.byte_size + quantization.bytes_size

                    return struct.pack('>I', patch_size) + tensor_data.bytes + quantization.bytes, 4 + patch_size

    return None



import re
from struct import pack
from typing import Generator


class Quantization:
    """
    Match quantization data in Edge Impulse model.
    Patch format is (for each layer):
        index: 1 byte
        size: 2 bytes
        data: size * 8 bytes
            scale: 4 bytes (float)
            zero point: 4 bytes (int)
    """
    def __init__(self, contents: str):
        def eval_data(data: str, dtype) -> list:
            return [dtype(x) for x in eval("[" + re.sub(r'/\*.+?\*/', "", data) + "]")]

        scales = re.findall(r'TfArray<(\d+), float> quant(\d+)_scale = \{ \d+, \{([\s\S]+?)}', contents)
        zeros = re.findall(r'TfArray<(\d+), int> quant(\d+)_zero = \{ \d+, \{([\s\S]+?)}', contents)
        assert len(scales) > 0, "Can't find quantization data"

        self.scales = [(int(index), eval(size), eval_data(data, float)) for size, index, data in scales]
        self.zeros = [eval_data(data, int) for size, index, data in zeros]

        assert len(self.scales) == len(self.zeros), f"quantization data size mismatch ({len(self.scales)} scales vs {len(self.zeros)} zeros)"

    @property
    def bytes_size(self) -> int:
        return len(self.bytes)

    @property
    def bytes(self):
        """
        Convert data to binary format
        :return:
        """
        packs = []

        for (index, size, scale), zero in zip(self.scales, self.zeros):
            data = [x for pair in zip(scale, zero) for x in pair]
            packs.append(pack(f">BH{'fi' * len(scale)}", index, size, *data))

        return b"".join(packs)

    @property
    def patches(self) -> Generator:
        """
        Generate code to patch quantization data.
        :return:
        """
        for index, size, _ in self.scales:
            yield f"""if (!ei_patch_tensor_quantization(stream, {index}, {size})) return false;"""



import os.path
import re
from jinja2 import Environment, FileSystemLoader

from edgeimpulse_ota.InputLibrary import InputLibrary
from edgeimpulse_ota.TensorData import TensorData
from edgeimpulse_ota.Quantization import Quantization


def patch_header(contents: str) -> str:
    """
    Add OTA functions prototypes to header
    :param contents:
    :return:
    """
    return contents.replace("#endif", "bool ei_ota(Stream& stream);\nbool ei_restore();\n\n#endif")


def patch_cpp(contents: str) -> str:
    """
    Implement OTA update logic
    :param contents:
    :return:
    """
    env = Environment(loader=FileSystemLoader(os.path.join(os.path.dirname(__file__), "templates")))
    template = env.get_template("ota.jinja")

    tensor_data = TensorData(contents)
    quantization = Quantization(contents)

    # remove const from tensors' data and quantization
    contents = re.sub(r"const TfArray<(\d+), (float|int)> quant", "TfArray<\g<1>, \g<2>> quant", contents)
    contents = contents.replace("const ALIGN", "ALIGN")

    return template.render(
        contents=contents,
        tensor_data=tensor_data,
        quantization=quantization,
        patch_size=tensor_data.byte_size + quantization.bytes_size
    )


def patch(zip: str | bytes = None, api_key: str = None, project_id: str = None, **kwargs):
    """
    Apply patch to library
    :param zip: if str, it is interpreted as a path to a local file. If bytes, it is interpreted as the zip contents
    :param api_key: Edge Impulse API key
    :param project_id: Edge Impulse project ID
    :return:
    """
    library = InputLibrary(zip=zip, api_key=api_key, project_id=project_id, **kwargs)
    library.replace({
        "tflite-model/tflite_learn_*.cpp": patch_cpp,
        "tflite-model/tflite_learn_*.h": patch_header
    })

    return library.bytes



from distutils.core import setup


setup(
    name='edgeimpulse_ota',
    packages=['edgeimpulse_ota'],
    version='1.0.1',
    license='MIT',
    description='OTA updates for Edge Impulse models',
    author='Simone Salerno',
    author_email='support@eloquentarduino.com',
    url='https://github.com/eloquentarduino/python-edgeimpulse-ota',
    keywords=[
        'ML',
        'Edge AI'
    ],
    install_requires=[
        'Jinja2',
        'requests'
    ],
    package_data={
        'edgeimpulse_ota': ['templates/*.jinja']
    },
)
